"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = __importDefault(require("assert"));
const Module = require('module');
/**
 * Hooking a function means replacing it with our own implementation that (optionally)
 * delegates to the original.
 *
 * Multiple hooks should be allowed, delegating in a chain.
 *
 * If we install hook A, then install hook B, then uninstall A, hopefully this should not break.
 * However it requires that hook B appropriately skips over the now-uninstalled A.
 *
 * This manager wraps up that behavior.
 */
class HookManager {
    constructor() {
        /** Is our handler installed? */
        this._installed = false;
        // First item in this array delegates to the next.
        // last item delegates to the original implementation.
        this._hooks = [];
        const self = this;
        this._handler = function () {
            return self._hooks[0].fn.apply(this, arguments);
        };
    }
    /**
     * Install a hook.
     * Returns a function the hook can call to delegate to the next in the chain,
     * like a `super.foo()` call.
     */
    installHook(hookFn) {
        if (!this._installed) {
            this._installed = true;
            this._original = this.getFunction();
            this.setFunction(this._handler);
            // For simplicity, the original implementation is wrapped up as a hook
            const original = {
                fn: this._original,
                delegateTo: undefined
            };
            this._hooks.push(original);
        }
        const thisHook = {
            fn: hookFn,
            delegateTo: this._hooks[0]
        };
        this._hooks.unshift(thisHook);
        // will invoke the next hook
        return function delegator() {
            return thisHook.delegateTo.fn.apply(this, arguments);
        };
    }
    /**
     * Uninstall a previously installed hook.
     */
    uninstallHook(hookFn) {
        assert_1.default(this._installed, 'Must be installed before uninstalling');
        assert_1.default(hookFn !== this._original);
        const hookIndex = this._hooks.findIndex(h => h.fn === hookFn);
        assert_1.default(hookIndex >= 0);
        // splice out of the array
        this._hooks.splice(hookIndex, 1);
        // fix the preceding item's delegate to point to the next hook
        for (let i = 1; i < this._hooks.length; i++) {
            this._hooks[i - 1].delegateTo = this._hooks[i];
        }
        // If the only remaining hook fn is the one that represents `original`, then uninstall ourselves.
        if (this._hooks.length === 1) {
            this._installed = false;
            this.setFunction(this._original);
            this._hooks = [];
        }
    }
}
class ResolveFilenameHookManager extends HookManager {
    getFunction() {
        return Module._resolveFilename;
    }
    setFunction(fn) {
        Module._resolveFilename = fn;
    }
}
exports._resolveFilenameHookManager = new ResolveFilenameHookManager();
//# sourceMappingURL=core.js.map